@@ -1,4 +1,4 @@
-# Copyright (C) 2020-2021 Sebastian Blauth
+# Copyright (C) 2020-2022 Sebastian Blauth
 #
 # This file is part of CASHOCS.
 #
@@ -22,8 +22,14 @@ It includes the optimization problem, the solution algorithms and
 the line search needed for this.
 """
 
-from .cost_functional import ReducedCostFunctional
+from .control_optimization_algorithm import ControlOptimizationAlgorithm
+from .cost_functional import ReducedControlCostFunctional
 from .line_search import ArmijoLineSearch
-from .optimization_algorithm import OptimizationAlgorithm
+from .optimal_control_problem import OptimalControlProblem
 
-__all__ = ["ReducedCostFunctional", "ArmijoLineSearch", "OptimizationAlgorithm"]
+__all__ = [
+    "ReducedControlCostFunctional",
+    "ArmijoLineSearch",
+    "ControlOptimizationAlgorithm",
+    "OptimalControlProblem",
+]
