@@ -1,4 +1,4 @@
-# Copyright (C) 2020-2021 Sebastian Blauth
+# Copyright (C) 2020-2022 Sebastian Blauth
 #
 # This file is part of CASHOCS.
 #
@@ -19,23 +19,34 @@
 
 """
 
+from __future__ import annotations
+
+import configparser
+from typing import Dict, List, Union, Optional
+
 import fenics
 import numpy as np
 import ufl
+from typing_extensions import Literal
 
 from .methods import NCG, GradientDescent, LBFGS, Newton, PDAS
-from .._exceptions import ConfigError, InputError
+from .. import verification
+from .._exceptions import InputError
 from .._forms import ControlFormHandler
-from .._optimal_control import ReducedCostFunctional
+from .._interfaces import OptimizationProblem
+from .._optimal_control import ReducedControlCostFunctional
 from .._pde_problems import (
     AdjointProblem,
-    GradientProblem,
+    ControlGradientProblem,
     HessianProblem,
     StateProblem,
     UnconstrainedHessianProblem,
 )
-from ..optimization_problem import OptimizationProblem
-from ..utils import _optimization_algorithm_configuration, _check_and_enlist_functions
+from ..utils import (
+    _optimization_algorithm_configuration,
+    enlist,
+    _check_and_enlist_control_constraints,
+)
 
 
 class OptimalControlProblem(OptimizationProblem):
@@ -51,80 +62,139 @@ class OptimalControlProblem(OptimizationProblem):
     and so on.
     """
 
+    def __new__(
+        cls,
+        state_forms: Union[ufl.Form, List[ufl.Form]],
+        bcs_list: Union[
+            fenics.DirichletBC, List[fenics.DirichletBC], List[List[fenics.DirichletBC]]
+        ],
+        cost_functional_form: Union[ufl.Form, List[ufl.Form]],
+        states: Union[fenics.Function, List[fenics.Function]],
+        controls: Union[fenics.Function, List[fenics.Function]],
+        adjoints: Union[fenics.Function, List[fenics.Function]],
+        config: Optional[configparser.ConfigParser] = None,
+        riesz_scalar_products: Optional[Union[ufl.Form, List[ufl.Form]]] = None,
+        control_constraints: Optional[List[List[Union[float, fenics.Function]]]] = None,
+        initial_guess: Optional[List[fenics.Function]] = None,
+        ksp_options: Optional[List[List[List[str]]]] = None,
+        adjoint_ksp_options: Optional[List[List[List[str]]]] = None,
+        scalar_tracking_forms: Optional[Union[Dict, List[Dict]]] = None,
+        min_max_terms: Optional[List[Dict]] = None,
+        desired_weights: Optional[List[float]] = None,
+    ) -> OptimalControlProblem:
+
+        if desired_weights is not None:
+            _use_scaling = True
+        else:
+            _use_scaling = False
+
+        if _use_scaling:
+            unscaled_problem = super().__new__(cls)
+            unscaled_problem.__init__(
+                state_forms,
+                bcs_list,
+                cost_functional_form,
+                states,
+                controls,
+                adjoints,
+                config=config,
+                riesz_scalar_products=riesz_scalar_products,
+                control_constraints=control_constraints,
+                initial_guess=initial_guess,
+                ksp_options=ksp_options,
+                adjoint_ksp_options=adjoint_ksp_options,
+                scalar_tracking_forms=scalar_tracking_forms,
+                min_max_terms=min_max_terms,
+                desired_weights=desired_weights,
+            )
+            unscaled_problem._scale_cost_functional()  # overwrites the cost functional list
+
+        return super().__new__(cls)
+
     def __init__(
         self,
-        state_forms,
-        bcs_list,
-        cost_functional_form,
-        states,
-        controls,
-        adjoints,
-        config=None,
-        riesz_scalar_products=None,
-        control_constraints=None,
-        initial_guess=None,
-        ksp_options=None,
-        adjoint_ksp_options=None,
-        desired_weights=None,
-        scalar_tracking_forms=None,
-    ):
+        state_forms: Union[ufl.Form, List[ufl.Form]],
+        bcs_list: Union[
+            fenics.DirichletBC, List[fenics.DirichletBC], List[List[fenics.DirichletBC]]
+        ],
+        cost_functional_form: Union[ufl.Form, List[ufl.Form]],
+        states: Union[fenics.Function, List[fenics.Function]],
+        controls: Union[fenics.Function, List[fenics.Function]],
+        adjoints: Union[fenics.Function, List[fenics.Function]],
+        config: Optional[configparser.ConfigParser] = None,
+        riesz_scalar_products: Optional[Union[ufl.Form, List[ufl.Form]]] = None,
+        control_constraints: Optional[List[List[Union[float, fenics.Function]]]] = None,
+        initial_guess: Optional[List[fenics.Function]] = None,
+        ksp_options: Optional[List[List[List[str]]]] = None,
+        adjoint_ksp_options: Optional[List[List[List[str]]]] = None,
+        scalar_tracking_forms: Optional[Union[Dict, List[Dict]]] = None,
+        min_max_terms: Optional[List[Dict]] = None,
+        desired_weights: Optional[List[float]] = None,
+    ) -> None:
         r"""This is used to generate all classes and functionalities. First ensures
         consistent input, afterwards, the solution algorithm is initialized.
 
         Parameters
         ----------
         state_forms : ufl.form.Form or list[ufl.form.Form]
-                The weak form of the state equation (user implemented). Can be either
-                a single UFL form, or a (ordered) list of UFL forms.
-        bcs_list : list[dolfin.fem.dirichletbc.DirichletBC] or list[list[dolfin.fem.dirichletbc.DirichletBC]] or dolfin.fem.dirichletbc.DirichletBC or None
-                The list of DirichletBC objects describing Dirichlet (essential) boundary conditions.
-                If this is ``None``, then no Dirichlet boundary conditions are imposed.
+            The weak form of the state equation (user implemented). Can be either
+            a single UFL form, or a (ordered) list of UFL forms.
+        bcs_list : list[fenics.DirichletBC] or list[list[fenics.DirichletBC]] or fenics.DirichletBC or None
+            The list of DirichletBC objects describing Dirichlet (essential) boundary conditions.
+            If this is ``None``, then no Dirichlet boundary conditions are imposed.
         cost_functional_form : ufl.form.Form or list[ufl.form.Form]
-                UFL form of the cost functional.
-        states : dolfin.function.function.Function or list[dolfin.function.function.Function]
-                The state variable(s), can either be a :py:class:`fenics.Function`, or a list of these.
-        controls : dolfin.function.function.Function or list[dolfin.function.function.Function]
-                The control variable(s), can either be a :py:class:`fenics.Function`, or a list of these.
-        adjoints : dolfin.function.function.Function or list[dolfin.function.function.Function]
-                The adjoint variable(s), can either be a :py:class:`fenics.Function`, or a (ordered) list of these.
-        config : configparser.ConfigParser or None
-                The config file for the problem, generated via :py:func:`cashocs.create_config`.
-                Alternatively, this can also be ``None``, in which case the default configurations
-                are used, except for the optimization algorithm. This has then to be specified
-                in the :py:meth:`solve <cashocs.OptimalControlProblem.solve>` method. The
-                default is ``None``.
+            UFL form of the cost functional.
+        states : fenics.Function or list[fenics.Function]
+            The state variable(s), can either be a :py:class:`fenics.Function`, or a list of these.
+        controls : fenics.Function or list[fenics.Function]
+            The control variable(s), can either be a :py:class:`fenics.Function`, or a list of these.
+        adjoints : fenics.Function or list[fenics.Function]
+            The adjoint variable(s), can either be a :py:class:`fenics.Function`, or a (ordered) list of these.
+        config : configparser.ConfigParser or None, optional
+            The config file for the problem, generated via :py:func:`cashocs.create_config`.
+            Alternatively, this can also be ``None``, in which case the default configurations
+            are used, except for the optimization algorithm. This has then to be specified
+            in the :py:meth:`solve <cashocs.OptimalControlProblem.solve>` method. The
+            default is ``None``.
         riesz_scalar_products : None or ufl.form.Form or list[ufl.form.Form], optional
-                The scalar products of the control space. Can either be None, a single UFL form, or a
-                (ordered) list of UFL forms. If ``None``, the :math:`L^2(\Omega)` product is used.
-                (default is ``None``).
-        control_constraints : None or list[dolfin.function.function.Function] or list[float] or list[list[dolfin.function.function.Function]] or list[list[float]], optional
-                Box constraints posed on the control, ``None`` means that there are none (default is ``None``).
-                The (inner) lists should contain two elements of the form ``[u_a, u_b]``, where ``u_a`` is the lower,
-                and ``u_b`` the upper bound.
-        initial_guess : list[dolfin.function.function.Function], optional
-                List of functions that act as initial guess for the state variables, should be valid input for :py:func:`fenics.assign`.
-                Defaults to ``None``, which means a zero initial guess.
+            The scalar products of the control space. Can either be None, a single UFL form, or a
+            (ordered) list of UFL forms. If ``None``, the :math:`L^2(\Omega)` product is used.
+            (default is ``None``).
+        control_constraints : None or list[fenics.Function] or list[float] or list[list[fenics.Function]] or list[list[float]], optional
+            Box constraints posed on the control, ``None`` means that there are none (default is ``None``).
+            The (inner) lists should contain two elements of the form ``[u_a, u_b]``, where ``u_a`` is the lower,
+            and ``u_b`` the upper bound.
+        initial_guess : list[fenics.Function], optional
+            List of functions that act as initial guess for the state variables, should be valid input for :py:func:`fenics.assign`.
+            Defaults to ``None``, which means a zero initial guess.
         ksp_options : list[list[str]] or list[list[list[str]]] or None, optional
-                A list of strings corresponding to command line options for PETSc,
-                used to solve the state systems. If this is ``None``, then the direct solver
-                mumps is used (default is ``None``).
-        adjoint_ksp_options : list[list[str]] or list[list[list[str]]] or None
-                A list of strings corresponding to command line options for PETSc,
-                used to solve the adjoint systems. If this is ``None``, then the same options
-                as for the state systems are used (default is ``None``).
-        scalar_tracking_forms : dict or list[dict]
-                A list of dictionaries that define scalar tracking type cost functionals,
-                where an integral value should be brought to a desired value. Each dict needs
-                to have the keys ``'integrand'`` and ``'tracking_goal'``. Default is ``None``,
-                i.e., no scalar tracking terms are considered.
+            A list of strings corresponding to command line options for PETSc,
+            used to solve the state systems. If this is ``None``, then the direct solver
+            mumps is used (default is ``None``).
+        adjoint_ksp_options : list[list[str]] or list[list[list[str]]] or None, optional
+            A list of strings corresponding to command line options for PETSc,
+            used to solve the adjoint systems. If this is ``None``, then the same options
+            as for the state systems are used (default is ``None``).
+        scalar_tracking_forms : dict or list[dict] or None, optional
+            A list of dictionaries that define scalar tracking type cost functionals,
+            where an integral value should be brought to a desired value. Each dict needs
+            to have the keys ``'integrand'`` and ``'tracking_goal'``. Default is ``None``,
+            i.e., no scalar tracking terms are considered.
+        min_max_terms : dict or list[dict] or None, optional
+            Additional terms for the cost functional, not to be used directly.
+        desired_weights : list[float] or None, optional
+            A list of values for scaling the cost functional terms. If this is supplied,
+            the cost functional has to be given as list of summands. The individual terms
+            are then scaled, so that term `i` has the magnitude of `desired_weights[i]`
+            for the initial iteration. In case that `desired_weights` is `None`, no scaling
+            is performed. Default is `None`.
 
         Examples
         --------
         Examples how to use this class can be found in the :ref:`tutorial <tutorial_index>`.
         """
 
-        OptimizationProblem.__init__(
-            self,
+        super().__init__(
             state_forms,
             bcs_list,
             cost_functional_form,
@@ -134,19 +204,12 @@ class OptimalControlProblem(OptimizationProblem):
             initial_guess,
             ksp_options,
             adjoint_ksp_options,
-            desired_weights,
             scalar_tracking_forms,
+            min_max_terms,
+            desired_weights,
         )
 
-        try:
-            self.controls = _check_and_enlist_functions(controls)
-        except InputError:
-            raise InputError(
-                "cashocs._optimal_control.optimal_control_problem.OptimalControlProblem",
-                "controls",
-                "Type of controls is wrong.",
-            )
-
+        self.controls = enlist(controls)
         self.control_dim = len(self.controls)
 
         ### riesz_scalar_products
@@ -161,101 +224,22 @@ class OptimalControlProblem(OptimizationProblem):
                 for i in range(len(self.controls))
             ]
         else:
-            try:
-                if (
-                    isinstance(riesz_scalar_products, list)
-                    and len(riesz_scalar_products) > 0
-                ):
-                    for i in range(len(riesz_scalar_products)):
-                        if isinstance(riesz_scalar_products[i], ufl.form.Form):
-                            pass
-                        else:
-                            raise InputError(
-                                "cashocs._optimal_control.optimal_control_problem.OptimalControlProblem",
-                                "riesz_scalar_products",
-                                "riesz_scalar_products have to be ufl forms",
-                            )
-                    self.riesz_scalar_products = riesz_scalar_products
-                elif isinstance(riesz_scalar_products, ufl.form.Form):
-                    self.riesz_scalar_products = [riesz_scalar_products]
-                else:
-                    raise InputError(
-                        "cashocs._optimal_control.optimal_control_problem.OptimalControlProblem",
-                        "riesz_scalar_products",
-                        "riesz_scalar_products have to be ufl forms",
-                    )
-            except:
-                raise InputError(
-                    "cashocs._optimal_control.optimal_control_problem.OptimalControlProblem",
-                    "riesz_scalar_products",
-                    "riesz_scalar_products have to be ufl forms",
-                )
+            self.riesz_scalar_products = enlist(riesz_scalar_products)
+            self.uses_custom_scalar_product = True
 
         ### control_constraints
         if control_constraints is None:
             self.control_constraints = []
             for control in self.controls:
                 u_a = fenics.Function(control.function_space())
-                u_a.vector()[:] = float("-inf")
+                u_a.vector().vec().set(float("-inf"))
                 u_b = fenics.Function(control.function_space())
-                u_b.vector()[:] = float("inf")
+                u_b.vector().vec().set(float("inf"))
                 self.control_constraints.append([u_a, u_b])
         else:
-            try:
-                if (
-                    isinstance(control_constraints, list)
-                    and len(control_constraints) > 0
-                ):
-                    if isinstance(control_constraints[0], list):
-                        for i in range(len(control_constraints)):
-                            if (
-                                isinstance(control_constraints[i], list)
-                                and len(control_constraints[i]) == 2
-                            ):
-                                for j in range(2):
-                                    if isinstance(
-                                        control_constraints[i][j], (float, int)
-                                    ):
-                                        pass
-                                    elif isinstance(
-                                        control_constraints[i][j], fenics.Function
-                                    ):
-                                        pass
-                                    else:
-                                        raise InputError(
-                                            "cashocs._optimal_control.optimal_control_problem.OptimalControlProblem",
-                                            "control_constraints",
-                                            "control_constraints has to be a list containing upper and lower bounds",
-                                        )
-                            else:
-                                raise InputError(
-                                    "cashocs._optimal_control.optimal_control_problem.OptimalControlProblem",
-                                    "control_constraints",
-                                    "control_constraints has to be a list containing upper and lower bounds",
-                                )
-                        self.control_constraints = control_constraints
-                    elif (
-                        isinstance(control_constraints[0], (float, int))
-                        or (isinstance(control_constraints, fenics.Function))
-                    ) and (
-                        isinstance(control_constraints[1], (float, int))
-                        or (isinstance(control_constraints[1], fenics.Function))
-                    ):
-
-                        self.control_constraints = [control_constraints]
-                    else:
-                        raise InputError(
-                            "cashocs._optimal_control.optimal_control_problem.OptimalControlProblem",
-                            "control_constraints",
-                            "control_constraints has to be a list containing upper and lower bounds",
-                        )
-
-            except:
-                raise InputError(
-                    "cashocs._optimal_control.optimal_control_problem.OptimalControlProblem",
-                    "control_constraints",
-                    "control_constraints has to be a list containing upper and lower bounds",
-                )
+            self.control_constraints = _check_and_enlist_control_constraints(
+                control_constraints
+            )
 
         # recast floats into functions for compatibility
         temp_constraints = self.control_constraints[:]
@@ -263,7 +247,7 @@ class OptimalControlProblem(OptimizationProblem):
         for idx, pair in enumerate(temp_constraints):
             if isinstance(pair[0], (float, int)):
                 lower_bound = fenics.Function(self.controls[idx].function_space())
-                lower_bound.vector()[:] = pair[0]
+                lower_bound.vector().vec().set(pair[0])
             elif isinstance(pair[0], fenics.Function):
                 lower_bound = pair[0]
             else:
@@ -275,7 +259,7 @@ class OptimalControlProblem(OptimizationProblem):
 
             if isinstance(pair[1], (float, int)):
                 upper_bound = fenics.Function(self.controls[idx].function_space())
-                upper_bound.vector()[:] = pair[1]
+                upper_bound.vector().vec().set(pair[1])
             elif isinstance(pair[1], fenics.Function):
                 upper_bound = pair[1]
             else:
@@ -297,9 +281,9 @@ class OptimalControlProblem(OptimizationProblem):
                     "The lower bound must always be smaller than the upper bound for the control_constraints.",
                 )
 
-            if np.max(pair[0].vector()[:]) == float("-inf") and np.min(
-                pair[1].vector()[:]
-            ) == float("inf"):
+            if pair[0].vector().vec().max()[1] == float("-inf") and pair[
+                1
+            ].vector().vec().min()[1] == float("inf"):
                 # no control constraint for this component
                 pass
             else:
@@ -362,20 +346,7 @@ class OptimalControlProblem(OptimizationProblem):
             )
         ### end overloading
 
-        self.form_handler = ControlFormHandler(
-            self.lagrangian,
-            self.bcs_list,
-            self.states,
-            self.controls,
-            self.adjoints,
-            self.config,
-            self.riesz_scalar_products,
-            self.control_constraints,
-            self.ksp_options,
-            self.adjoint_ksp_options,
-            self.require_control_constraints,
-            use_scalar_tracking=self.use_scalar_tracking,
-        )
+        self.form_handler = ControlFormHandler(self)
 
         self.state_spaces = self.form_handler.state_spaces
         self.control_spaces = self.form_handler.control_spaces
@@ -385,20 +356,20 @@ class OptimalControlProblem(OptimizationProblem):
 
         self.state_problem = StateProblem(self.form_handler, self.initial_guess)
         self.adjoint_problem = AdjointProblem(self.form_handler, self.state_problem)
-        self.gradient_problem = GradientProblem(
+        self.gradient_problem = ControlGradientProblem(
             self.form_handler, self.state_problem, self.adjoint_problem
         )
 
         self.algorithm = _optimization_algorithm_configuration(self.config)
 
-        self.reduced_cost_functional = ReducedCostFunctional(
+        self.reduced_cost_functional = ReducedControlCostFunctional(
             self.form_handler, self.state_problem
         )
 
-        self.gradients = self.gradient_problem.gradients
+        self.gradient = self.gradient_problem.gradient
         self.objective_value = 1.0
 
-    def _erase_pde_memory(self):
+    def _erase_pde_memory(self) -> None:
         """Resets the memory of the PDE problems so that new solutions are computed.
 
         This sets the value of has_solution to False for all relevant PDE problems,
@@ -409,11 +380,29 @@ class OptimalControlProblem(OptimizationProblem):
         None
         """
 
-        self.state_problem.has_solution = False
-        self.adjoint_problem.has_solution = False
+        super()._erase_pde_memory()
         self.gradient_problem.has_solution = False
 
-    def solve(self, algorithm=None, rtol=None, atol=None, max_iter=None):
+    def solve(
+        self,
+        algorithm: Optional[
+            Literal[
+                "gradient_descent",
+                "gd",
+                "conjugate_gradient",
+                "nonlinear_cg",
+                "ncg",
+                "lbfgs",
+                "bfgs",
+                "newton",
+                "pdas",
+                "primal_dual_active_set",
+            ]
+        ] = None,
+        rtol: Optional[float] = None,
+        atol: Optional[float] = None,
+        max_iter: Optional[int] = None,
+    ) -> None:
         r"""Solves the optimization problem by the method specified in the config file.
 
         Updates / overwrites states, controls, and adjoints according
@@ -422,31 +411,31 @@ class OptimalControlProblem(OptimizationProblem):
         Parameters
         ----------
         algorithm : str or None, optional
-                Selects the optimization algorithm. Valid choices are
-                ``'gradient_descent'`` or ``'gd'`` for a gradient descent method,
-                ``'conjugate_gradient'``, ``'nonlinear_cg'``, ``'ncg'`` or ``'cg'``
-                for nonlinear conjugate gradient methods, ``'lbfgs'`` or ``'bfgs'`` for
-                limited memory BFGS methods, ``'newton'`` for a truncated Newton method,
-                and ``'pdas'`` or ``'primal_dual_active_set'`` for a
-                primal dual active set method. This overwrites
-                the value specified in the config file. If this is ``None``,
-                then the value in the config file is used. Default is
-                ``None``.
+            Selects the optimization algorithm. Valid choices are
+            ``'gradient_descent'`` or ``'gd'`` for a gradient descent method,
+            ``'conjugate_gradient'``, ``'nonlinear_cg'``, ``'ncg'`` or ``'cg'``
+            for nonlinear conjugate gradient methods, ``'lbfgs'`` or ``'bfgs'`` for
+            limited memory BFGS methods, ``'newton'`` for a truncated Newton method,
+            and ``'pdas'`` or ``'primal_dual_active_set'`` for a
+            primal dual active set method. This overwrites
+            the value specified in the config file. If this is ``None``,
+            then the value in the config file is used. Default is
+            ``None``.
         rtol : float or None, optional
-                The relative tolerance used for the termination criterion.
-                Overwrites the value specified in the config file. If this
-                is ``None``, the value from the config file is taken. Default
-                is ``None``.
+            The relative tolerance used for the termination criterion.
+            Overwrites the value specified in the config file. If this
+            is ``None``, the value from the config file is taken. Default
+            is ``None``.
         atol : float or None, optional
-                The absolute tolerance used for the termination criterion.
-                Overwrites the value specified in the config file. If this
-                is ``None``, the value from the config file is taken. Default
-                is ``None``.
+            The absolute tolerance used for the termination criterion.
+            Overwrites the value specified in the config file. If this
+            is ``None``, the value from the config file is taken. Default
+            is ``None``.
         max_iter : int or None, optional
-                The maximum number of iterations the optimization algorithm
-                can carry out before it is terminated. Overwrites the value
-                specified in the config file. If this is ``None``, the value from
-                the config file is taken. Default is ``None``.
+            The maximum number of iterations the optimization algorithm
+            can carry out before it is terminated. Overwrites the value
+            specified in the config file. If this is ``None``, the value from
+            the config file is taken. Default is ``None``.
 
         Returns
         -------
@@ -470,7 +459,7 @@ class OptimalControlProblem(OptimizationProblem):
           .. math:: || \nabla J(u_k) || \leq \texttt{atol} + \texttt{rtol} || \nabla J(u_0) ||.
         """
 
-        self.algorithm = _optimization_algorithm_configuration(self.config, algorithm)
+        super().solve(algorithm=algorithm, rtol=rtol, atol=atol, max_iter=max_iter)
 
         if self.algorithm == "newton" or (
             self.algorithm == "pdas"
@@ -487,21 +476,6 @@ class OptimalControlProblem(OptimizationProblem):
                 self.form_handler, self.gradient_problem
             )
 
-        if (rtol is not None) and (atol is None):
-            self.config.set("OptimizationRoutine", "rtol", str(rtol))
-            self.config.set("OptimizationRoutine", "atol", str(0.0))
-        elif (atol is not None) and (rtol is None):
-            self.config.set("OptimizationRoutine", "rtol", str(0.0))
-            self.config.set("OptimizationRoutine", "atol", str(atol))
-        elif (atol is not None) and (rtol is not None):
-            self.config.set("OptimizationRoutine", "rtol", str(rtol))
-            self.config.set("OptimizationRoutine", "atol", str(atol))
-
-        if max_iter is not None:
-            self.config.set("OptimizationRoutine", "maximum_iterations", str(max_iter))
-
-        self._check_for_custom_forms()
-
         if self.algorithm == "gradient_descent":
             self.solver = GradientDescent(self)
         elif self.algorithm == "lbfgs":
@@ -521,18 +495,11 @@ class OptimalControlProblem(OptimizationProblem):
                 "'gradient_descent' ('gd'), 'lbfgs' ('bfgs'), 'conjugate_gradient' ('cg'), "
                 "'newton', or 'primal_dual_active_set' ('pdas').",
             )
-        else:
-            raise ConfigError(
-                "OptimizationRoutine",
-                "algorithm",
-                "Not a valid input. Needs to be one "
-                "of 'gradient_descent' ('gd'), 'lbfgs' ('bfgs'), 'conjugate_gradient' ('cg'), 'newton', or 'primal_dual_active_set' ('pdas').",
-            )
 
         self.solver.run()
         self.solver.post_processing()
 
-    def compute_gradient(self):
+    def compute_gradient(self) -> List[fenics.Function]:
         """Solves the Riesz problem to determine the gradient.
 
         This can be used for debugging, or code validation.
@@ -541,24 +508,24 @@ class OptimalControlProblem(OptimizationProblem):
 
         Returns
         -------
-        list[dolfin.function.function.Function]
-                A list consisting of the (components) of the gradient.
+        list[fenics.Function]
+            A list consisting of the (components) of the gradient.
         """
 
         self.gradient_problem.solve()
 
-        return self.gradients
+        return self.gradient
 
-    def supply_derivatives(self, derivatives):
-        """Overrides the derivatives of the reduced cost functional w.r.t. controls.
+    def supply_derivatives(self, derivatives: Union[ufl.Form, List[ufl.Form]]) -> None:
+        """Overwrites the derivatives of the reduced cost functional w.r.t. controls.
 
         This allows users to implement their own derivatives and use cashocs as a
         solver library only.
 
         Parameters
         ----------
-        derivatives : ufl.form.Form or list[ufl.form.Form]
-                The derivatives of the reduced (!) cost functional w.r.t. controls.
+        derivatives : ufl.Form or list[ufl.Form]
+            The derivatives of the reduced (!) cost functional w.r.t. controls.
 
         Returns
         -------
@@ -626,7 +593,14 @@ class OptimalControlProblem(OptimizationProblem):
         self.form_handler.gradient_forms_rhs = mod_derivatives
         self.has_custom_derivative = True
 
-    def supply_custom_forms(self, derivatives, adjoint_forms, adjoint_bcs_list):
+    def supply_custom_forms(
+        self,
+        derivatives: Union[ufl.Form, List[ufl.Form]],
+        adjoint_forms: Union[ufl.Form, List[ufl.Form]],
+        adjoint_bcs_list: Union[
+            fenics.DirichletBC, List[fenics.DirichletBC], List[List[fenics.DirichletBC]]
+        ],
+    ) -> None:
         """Overrides both adjoint system and derivatives with user input.
 
         This allows the user to specify both the derivatives of the reduced cost functional
@@ -639,12 +613,12 @@ class OptimalControlProblem(OptimizationProblem):
 
         Parameters
         ----------
-        derivatives : ufl.form.Form or list[ufl.form.Form]
-                The derivatives of the reduced (!) cost functional w.r.t. controls.
-        adjoint_forms : ufl.form.Form or list[ufl.form.Form]
-                The UFL forms of the adjoint system(s).
-        adjoint_bcs_list : list[dolfin.fem.dirichletbc.DirichletBC] or list[list[dolfin.fem.dirichletbc.DirichletBC]] or dolfin.fem.dirichletbc.DirichletBC or None
-                The list of Dirichlet boundary conditions for the adjoint system(s).
+        derivatives : ufl.Form or list[ufl.Form]
+            The derivatives of the reduced (!) cost functional w.r.t. controls.
+        adjoint_forms : ufl.Form or list[ufl.Form]
+            The UFL forms of the adjoint system(s).
+        adjoint_bcs_list : list[fenics.DirichletBC] or list[list[fenics.DirichletBC]] or fenics.DirichletBC or None
+            The list of Dirichlet boundary conditions for the adjoint system(s).
 
         Returns
         -------
@@ -653,3 +627,32 @@ class OptimalControlProblem(OptimizationProblem):
 
         self.supply_derivatives(derivatives)
         self.supply_adjoint_forms(adjoint_forms, adjoint_bcs_list)
+
+    def gradient_test(
+        self,
+        u: Optional[List[fenics.Function]] = None,
+        h: Optional[List[fenics.Function]] = None,
+        rng: Optional[np.random.RandomState] = None,
+    ) -> float:
+        """Taylor test to verify that the computed gradient is correct for optimal control problems.
+
+        Parameters
+        ----------
+        u : list[fenics.Function] or None, optional
+            The point, at which the gradient shall be verified. If this is ``None``,
+            then the current controls of the optimization problem are used. Default is
+            ``None``.
+        h : list[fenics.Function] or None, optional
+            The direction(s) for the directional (Gateaux) derivative. If this is ``None``,
+            one random direction is chosen. Default is ``None``.
+        rng : numpy.random.RandomState or None, optional
+            A numpy random state for calculating a random direction
+
+        Returns
+        -------
+        float
+            The convergence order from the Taylor test. If this is (approximately) 2 or larger,
+             everything works as expected.
+        """
+
+        return verification.control_gradient_test(self, u, h, rng)
