@@ -1,4 +1,4 @@
-# Copyright (C) 2020-2021 Sebastian Blauth
+# Copyright (C) 2020-2022 Sebastian Blauth
 #
 # This file is part of CASHOCS.
 #
@@ -22,35 +22,46 @@ the "Hessian problems" occurring in the truncated Newton
 method.
 """
 
+from __future__ import annotations
+
+import abc
+from typing import TYPE_CHECKING, List, Optional
+
 import fenics
 import numpy as np
 from petsc4py import PETSc
 
-from .._exceptions import CashocsException, ConfigError, NotConvergedError
+from .._exceptions import CashocsException, NotConvergedError
 from .._loggers import debug
 from ..utils import _assemble_petsc_system, _setup_petsc_options, _solve_linear_problem
 
 
-class BaseHessianProblem:
+if TYPE_CHECKING:
+    from .._forms import ControlFormHandler
+    from .control_gradient_problem import ControlGradientProblem
+
+
+class BaseHessianProblem(abc.ABC):
     """Base class for derived Hessian problems."""
 
-    def __init__(self, form_handler, gradient_problem):
-        """Initializes self.
-
+    def __init__(
+        self, form_handler: ControlFormHandler, gradient_problem: ControlGradientProblem
+    ) -> None:
+        """
         Parameters
         ----------
-        form_handler : cashocs._forms.ControlFormHandler
-                The FormHandler object for the optimization problem.
-        gradient_problem : cashocs._pde_problems.GradientProblem
-                The GradientProblem object (this is needed for the computation
-                of the Hessian).
+        form_handler : ControlFormHandler
+            The FormHandler object for the optimization problem.
+        gradient_problem : ControlGradientProblem
+            The ControlGradientProblem object (this is needed for the computation
+            of the Hessian).
         """
 
         self.form_handler = form_handler
         self.gradient_problem = gradient_problem
 
         self.config = self.form_handler.config
-        self.gradients = self.gradient_problem.gradients
+        self.gradient = self.gradient_problem.gradient
 
         self.inner_newton = self.config.get("AlgoTNM", "inner_newton", fallback="cr")
         self.max_it_inner_newton = self.config.getint(
@@ -131,7 +142,9 @@ class BaseHessianProblem:
         for i, ksp in enumerate(self.ksps):
             ksp.setOperators(self.form_handler.riesz_projection_matrices[i])
 
-    def hessian_application(self, h, out):
+    def hessian_application(
+        self, h: List[fenics.Function], out: List[fenics.Function]
+    ) -> None:
         r"""Computes the application of the Hessian to some element
 
         This is needed in the truncated Newton method where we solve the system
@@ -142,10 +155,10 @@ class BaseHessianProblem:
 
         Parameters
         ----------
-        h : list[dolfin.function.function.Function]
-                A function to which we want to apply the Hessian to.
-        out : list[dolfin.function.function.Function]
-                A list of functions into which the result is saved.
+        h : list[fenics.Function]
+            A function to which we want to apply the Hessian to.
+        out : list[fenics.Function]
+            A list of functions into which the result is saved.
 
         Returns
         -------
@@ -153,7 +166,7 @@ class BaseHessianProblem:
         """
 
         for i in range(self.control_dim):
-            self.test_directions[i].vector()[:] = h[i].vector()[:]
+            self.test_directions[i].vector().vec().aypx(0.0, h[i].vector().vec())
 
         self.states_prime = self.form_handler.states_prime
         self.adjoints_prime = self.form_handler.adjoints_prime
@@ -303,55 +316,109 @@ class BaseHessianProblem:
 
         self.no_sensitivity_solves += 2
 
-    def newton_solve(self, idx_active=None):
+    def newton_solve(
+        self, idx_active: Optional[List[int]] = None
+    ) -> List[fenics.Function]:
+        """Solves the problem with a truncated Newton method.
+
+        Parameters
+        ----------
+        idx_active : list[int] or None, optional
+            List of active indices
+
+        Returns
+        -------
+        list[fenics.Function]
+            The Newton increment
+
+        """
 
         self.gradient_problem.solve()
         self.form_handler.compute_active_sets()
 
         for j in range(self.control_dim):
-            self.delta_control[j].vector()[:] = 0.0
+            self.delta_control[j].vector().vec().set(0.0)
 
         if self.inner_newton == "cg":
             self.cg(idx_active)
         elif self.inner_newton == "cr":
             self.cr(idx_active)
-        else:
-            raise ConfigError(
-                "AlgoTNM",
-                "inner_newton",
-                "Not a valid choice. Needs to be either 'cg' or 'cr'.",
-            )
 
         return self.delta_control
 
-    def cg(self, idx_active=None):
+    @abc.abstractmethod
+    def cg(self, idx_active: Optional[List[int]] = None) -> None:
+        """Solves the (truncated) Newton step with a CG method
+
+        Parameters
+        ----------
+        idx_active : list[int] or None, optional
+            The list of active indices
+
+        Returns
+        -------
+        None
+        """
+
         pass
 
-    def cr(self, idx_active=None):
+    @abc.abstractmethod
+    def cr(self, idx_active: Optional[List[int]] = None) -> None:
+        """Solves the (truncated) Newton step with a CR method
+
+        Parameters
+        ----------
+        idx_active : list[int] or None, optional
+            The list of active indices
+
+        Returns
+        -------
+        None
+        """
+
         pass
 
 
 class HessianProblem(BaseHessianProblem):
     """PDE Problem used to solve the (reduced) Hessian problem."""
 
-    def __init__(self, form_handler, gradient_problem):
+    def __init__(
+        self, form_handler: ControlFormHandler, gradient_problem: ControlGradientProblem
+    ) -> None:
         """Initializes self.
 
         Parameters
         ----------
-        form_handler : cashocs._forms.ControlFormHandler
-                The FormHandler object for the optimization problem.
-        gradient_problem : cashocs._pde_problems.GradientProblem
-                The GradientProblem object (this is needed for the computation
-                of the Hessian).
+        form_handler : ControlFormHandler
+            The FormHandler object for the optimization problem.
+        gradient_problem : ControlGradientProblem
+            The ControlGradientProblem object (this is needed for the computation
+            of the Hessian).
         """
 
-        BaseHessianProblem.__init__(self, form_handler, gradient_problem)
+        super().__init__(form_handler, gradient_problem)
 
-    def reduced_hessian_application(self, h, out):
+    def reduced_hessian_application(
+        self, h: List[fenics.Function], out: List[fenics.Function]
+    ) -> None:
+        """Computes the application of the reduced Hessian on a direction.
+
+        This is needed to solve the Newton step with iterative solvers.
+
+        Parameters
+        ----------
+        h : list[fenics.Function]
+            The direction, onto which the reduced Hessian is applied
+        out : list[fenics.Function]
+            The output of the application of the (linear) operator
+
+        Returns
+        -------
+        None
+        """
 
         for j in range(self.control_dim):
-            out[j].vector()[:] = 0.0
+            out[j].vector().vec().set(0.0)
 
         self.form_handler.restrict_to_inactive_set(h, self.inactive_part)
         self.hessian_application(self.inactive_part, self.hessian_actions)
@@ -361,22 +428,51 @@ class HessianProblem(BaseHessianProblem):
         self.form_handler.restrict_to_active_set(h, self.active_part)
 
         for j in range(self.control_dim):
-            out[j].vector()[:] = (
-                self.active_part[j].vector()[:] + self.inactive_part[j].vector()[:]
+            out[j].vector().vec().aypx(
+                0.0,
+                self.active_part[j].vector().vec()
+                + self.inactive_part[j].vector().vec(),
             )
 
-    def newton_solve(self, idx_active=None):
+    def newton_solve(
+        self, idx_active: Optional[List[List[int]]] = None
+    ) -> List[fenics.Function]:
+        """Solves the Newton step with an iterative method
+
+        Parameters
+        ----------
+        idx_active : list[list[int]] or None, optional
+            The list of active indices
+
+        Returns
+        -------
+        list[fenics.Function]
+            The Newton increment
+        """
 
         if idx_active is not None:
             raise CashocsException("Must not pass idx_active to HessianProblem.")
 
-        return BaseHessianProblem.newton_solve(self)
+        return super().newton_solve()
 
-    def cg(self, idx_active=None):
+    def cg(self, idx_active: Optional[List[int]] = None) -> None:
+        """Solves the (truncated) Newton step with a CG method
+
+        Parameters
+        ----------
+        idx_active : list[int] or None, optional
+            The list of active indices
+
+        Returns
+        -------
+        list[fenics.Function]
+            The Newton increment
+
+        """
 
         for j in range(self.control_dim):
-            self.residual[j].vector()[:] = -self.gradients[j].vector()[:]
-            self.p[j].vector()[:] = self.residual[j].vector()[:]
+            self.residual[j].vector().vec().aypx(0.0, -self.gradient[j].vector().vec())
+            self.p[j].vector().vec().aypx(0.0, self.residual[j].vector().vec())
 
         self.rsold = self.form_handler.scalar_product(self.residual, self.residual)
         self.eps_0 = np.sqrt(self.rsold)
@@ -395,8 +491,12 @@ class HessianProblem(BaseHessianProblem):
             self.alpha = self.rsold / sp_val
 
             for j in range(self.control_dim):
-                self.delta_control[j].vector()[:] += self.alpha * self.p[j].vector()[:]
-                self.residual[j].vector()[:] -= self.alpha * self.q[j].vector()[:]
+                self.delta_control[j].vector().vec().axpy(
+                    self.alpha, self.p[j].vector().vec()
+                )
+                self.residual[j].vector().vec().axpy(
+                    -self.alpha, self.q[j].vector().vec()
+                )
 
             self.rsnew = self.form_handler.scalar_product(self.residual, self.residual)
             self.eps = np.sqrt(self.rsnew)
@@ -407,17 +507,28 @@ class HessianProblem(BaseHessianProblem):
             self.beta = self.rsnew / self.rsold
 
             for j in range(self.control_dim):
-                self.p[j].vector()[:] = (
-                    self.residual[j].vector()[:] + self.beta * self.p[j].vector()[:]
+                self.p[j].vector().vec().aypx(
+                    self.beta, self.residual[j].vector().vec()
                 )
 
             self.rsold = self.rsnew
 
-    def cr(self, idx_active=None):
+    def cr(self, idx_active: Optional[List[int]] = None) -> None:
+        """Solves the (truncated) Newton step with a CR method
+
+        Parameters
+        ----------
+        idx_active : list[int] or None, optional
+            The list of active indices
+
+        Returns
+        -------
+        None
+        """
 
         for j in range(self.control_dim):
-            self.residual[j].vector()[:] = -self.gradients[j].vector()[:]
-            self.p[j].vector()[:] = self.residual[j].vector()[:]
+            self.residual[j].vector().vec().aypx(0.0, -self.gradient[j].vector().vec())
+            self.p[j].vector().vec().aypx(0.0, self.residual[j].vector().vec())
 
         self.eps_0 = np.sqrt(
             self.form_handler.scalar_product(self.residual, self.residual)
@@ -426,7 +537,7 @@ class HessianProblem(BaseHessianProblem):
         self.reduced_hessian_application(self.residual, self.s)
 
         for j in range(self.control_dim):
-            self.q[j].vector()[:] = self.s[j].vector()[:]
+            self.q[j].vector().vec().aypx(0.0, self.s[j].vector().vec())
 
         self.form_handler.restrict_to_active_set(self.residual, self.temp1)
         self.form_handler.restrict_to_active_set(self.s, self.temp2)
@@ -448,8 +559,12 @@ class HessianProblem(BaseHessianProblem):
             self.alpha = self.rAr / denominator
 
             for j in range(self.control_dim):
-                self.delta_control[j].vector()[:] += self.alpha * self.p[j].vector()[:]
-                self.residual[j].vector()[:] -= self.alpha * self.q[j].vector()[:]
+                self.delta_control[j].vector().vec().axpy(
+                    self.alpha, self.p[j].vector().vec()
+                )
+                self.residual[j].vector().vec().axpy(
+                    -self.alpha, self.q[j].vector().vec()
+                )
 
             self.eps = np.sqrt(
                 self.form_handler.scalar_product(self.residual, self.residual)
@@ -474,12 +589,10 @@ class HessianProblem(BaseHessianProblem):
             self.beta = self.rAr_new / self.rAr
 
             for j in range(self.control_dim):
-                self.p[j].vector()[:] = (
-                    self.residual[j].vector()[:] + self.beta * self.p[j].vector()[:]
-                )
-                self.q[j].vector()[:] = (
-                    self.s[j].vector()[:] + self.beta * self.q[j].vector()[:]
+                self.p[j].vector().vec().aypx(
+                    self.beta, self.residual[j].vector().vec()
                 )
+                self.q[j].vector().vec().aypx(self.beta, self.s[j].vector().vec())
 
             self.rAr = self.rAr_new
 
@@ -487,30 +600,54 @@ class HessianProblem(BaseHessianProblem):
 class UnconstrainedHessianProblem(BaseHessianProblem):
     """Hessian Problem without control constraints for the inner solver in PDAS."""
 
-    def __init__(self, form_handler, gradient_problem):
-        """Initializes self.
-
+    def __init__(
+        self, form_handler: ControlFormHandler, gradient_problem: ControlGradientProblem
+    ) -> None:
+        """
         Parameters
         ----------
-        form_handler : cashocs._forms.ControlFormHandler
-                The FormHandler object for the optimization problem.
-        gradient_problem : cashocs._pde_problems.GradientProblem
-                The GradientProblem object (this is needed for the computation
-                of the Hessian).
+        form_handler : ControlFormHandler
+            The FormHandler object for the optimization problem.
+        gradient_problem : ControlGradientProblem
+            The ControlGradientProblem object (this is needed for the computation
+            of the Hessian).
         """
 
-        BaseHessianProblem.__init__(self, form_handler, gradient_problem)
+        super().__init__(form_handler, gradient_problem)
 
         self.reduced_gradient = [
             fenics.Function(self.form_handler.control_spaces[j])
-            for j in range(len(self.gradients))
+            for j in range(len(self.gradient))
         ]
         self.temp = [fenics.Function(V) for V in self.form_handler.control_spaces]
 
-    def reduced_hessian_application(self, h, out, idx_active):
+    def reduced_hessian_application(
+        self,
+        h: List[fenics.Function],
+        out: List[fenics.Function],
+        idx_active: List[int],
+    ) -> None:
+        """Computes the application of the reduced Hessian on a direction.
+
+        This is needed to solve the Newton step with iterative solvers.
+
+        Parameters
+        ----------
+        h : list[fenics.Function]
+            The direction, onto which the reduced Hessian is applied
+        out : list[fenics.Function]
+            The output of the application of the (linear) operator
+        idx_active : list[int]
+            The list of active indices
+
+        Returns
+        -------
+        None
+
+        """
 
         for j in range(self.control_dim):
-            self.temp[j].vector()[:] = h[j].vector()[:]
+            self.temp[j].vector().vec().aypx(0.0, h[j].vector().vec())
             self.temp[j].vector()[idx_active[j]] = 0.0
 
         self.hessian_application(self.temp, out)
@@ -518,7 +655,22 @@ class UnconstrainedHessianProblem(BaseHessianProblem):
         for j in range(self.control_dim):
             out[j].vector()[idx_active[j]] = 0.0
 
-    def newton_solve(self, idx_active=None):
+    def newton_solve(
+        self, idx_active: Optional[List[List[int]]] = None
+    ) -> List[fenics.Function]:
+        """Solves the Newton step with an iterative method
+
+        Parameters
+        ----------
+        idx_active : list[list[int]] or None, optional
+            The list of active indices
+
+        Returns
+        -------
+        list[fenics.Function]
+            The Newton increment
+
+        """
 
         if idx_active is None:
             raise CashocsException(
@@ -528,16 +680,31 @@ class UnconstrainedHessianProblem(BaseHessianProblem):
         self.gradient_problem.solve()
 
         for j in range(self.control_dim):
-            self.reduced_gradient[j].vector()[:] = self.gradients[j].vector()[:]
+            self.reduced_gradient[j].vector().vec().aypx(
+                0.0, self.gradient[j].vector().vec()
+            )
             self.reduced_gradient[j].vector()[idx_active[j]] = 0.0
 
-        return BaseHessianProblem.newton_solve(self, idx_active)
+        return super().newton_solve(idx_active)
 
-    def cg(self, idx_active=None):
+    def cg(self, idx_active: Optional[List[int]] = None) -> None:
+        """Solves the (truncated) Newton step with a CG method
+
+        Parameters
+        ----------
+        idx_active : list[int] or None, optional
+            The list of active indices
+
+        Returns
+        -------
+        None
+        """
 
         for j in range(self.control_dim):
-            self.residual[j].vector()[:] = -self.reduced_gradient[j].vector()[:]
-            self.p[j].vector()[:] = self.residual[j].vector()[:]
+            self.residual[j].vector().vec().aypx(
+                0.0, -self.reduced_gradient[j].vector().vec()
+            )
+            self.p[j].vector().vec().aypx(0.0, self.residual[j].vector().vec())
 
         self.rsold = self.form_handler.scalar_product(self.residual, self.residual)
         self.eps_0 = np.sqrt(self.rsold)
@@ -547,8 +714,12 @@ class UnconstrainedHessianProblem(BaseHessianProblem):
 
             self.alpha = self.rsold / self.form_handler.scalar_product(self.p, self.q)
             for j in range(self.control_dim):
-                self.delta_control[j].vector()[:] += self.alpha * self.p[j].vector()[:]
-                self.residual[j].vector()[:] -= self.alpha * self.q[j].vector()[:]
+                self.delta_control[j].vector().vec().axpy(
+                    self.alpha, self.p[j].vector().vec()
+                )
+                self.residual[j].vector().vec().axpy(
+                    -self.alpha, self.q[j].vector().vec()
+                )
 
             self.rsnew = self.form_handler.scalar_product(self.residual, self.residual)
             self.eps = np.sqrt(self.rsnew)
@@ -559,17 +730,30 @@ class UnconstrainedHessianProblem(BaseHessianProblem):
             self.beta = self.rsnew / self.rsold
 
             for j in range(self.control_dim):
-                self.p[j].vector()[:] = (
-                    self.residual[j].vector()[:] + self.beta * self.p[j].vector()[:]
+                self.p[j].vector().vec().aypx(
+                    self.beta, self.residual[j].vector().vec()
                 )
 
             self.rsold = self.rsnew
 
-    def cr(self, idx_active=None):
+    def cr(self, idx_active: Optional[List[int]] = None) -> None:
+        """Solves the (truncated) Newton step with a CR method
+
+        Parameters
+        ----------
+        idx_active : list[int] or None, optional
+            The list of active indices
+
+        Returns
+        -------
+        None
+        """
 
         for j in range(self.control_dim):
-            self.residual[j].vector()[:] = -self.reduced_gradient[j].vector()[:]
-            self.p[j].vector()[:] = self.residual[j].vector()[:]
+            self.residual[j].vector().vec().aypx(
+                0.0, -self.reduced_gradient[j].vector().vec()
+            )
+            self.p[j].vector().vec().aypx(0.0, self.residual[j].vector().vec())
 
         self.eps_0 = np.sqrt(
             self.form_handler.scalar_product(self.residual, self.residual)
@@ -578,7 +762,7 @@ class UnconstrainedHessianProblem(BaseHessianProblem):
         self.reduced_hessian_application(self.residual, self.s, idx_active)
 
         for j in range(self.control_dim):
-            self.q[j].vector()[:] = self.s[j].vector()[:]
+            self.q[j].vector().vec().aypx(0.0, self.s[j].vector().vec())
 
         self.rAr = self.form_handler.scalar_product(self.residual, self.s)
 
@@ -586,8 +770,12 @@ class UnconstrainedHessianProblem(BaseHessianProblem):
             self.alpha = self.rAr / self.form_handler.scalar_product(self.q, self.q)
 
             for j in range(self.control_dim):
-                self.delta_control[j].vector()[:] += self.alpha * self.p[j].vector()[:]
-                self.residual[j].vector()[:] -= self.alpha * self.q[j].vector()[:]
+                self.delta_control[j].vector().vec().axpy(
+                    self.alpha, self.p[j].vector().vec()
+                )
+                self.residual[j].vector().vec().axpy(
+                    -self.alpha, self.q[j].vector().vec()
+                )
 
             self.eps = np.sqrt(
                 self.form_handler.scalar_product(self.residual, self.residual)
@@ -605,11 +793,9 @@ class UnconstrainedHessianProblem(BaseHessianProblem):
             self.beta = self.rAr_new / self.rAr
 
             for j in range(self.control_dim):
-                self.p[j].vector()[:] = (
-                    self.residual[j].vector()[:] + self.beta * self.p[j].vector()[:]
-                )
-                self.q[j].vector()[:] = (
-                    self.s[j].vector()[:] + self.beta * self.q[j].vector()[:]
+                self.p[j].vector().vec().aypx(
+                    self.beta, self.residual[j].vector().vec()
                 )
+                self.q[j].vector().vec().aypx(self.beta, self.s[j].vector().vec())
 
             self.rAr = self.rAr_new
