@@ -1,4 +1,4 @@
-# Copyright (C) 2020-2021 Sebastian Blauth
+# Copyright (C) 2020-2022 Sebastian Blauth
 #
 # This file is part of CASHOCS.
 #
@@ -19,27 +19,33 @@
 
 """
 
+from __future__ import annotations
+
+import configparser
 import json
 import os
 import subprocess
 import sys
 import tempfile
+from typing import Dict, List, Union, Optional
 
+import dolfin.function.argument
 import fenics
 import numpy as np
 import ufl
+from typing_extensions import Literal
 from ufl import replace
 from ufl.algorithms.estimate_degrees import estimate_total_polynomial_degree
 
 from .methods import NCG, GradientDescent, LBFGS
-from .._exceptions import CashocsException, ConfigError, InputError
+from .. import verification
+from .._exceptions import CashocsException, InputError
 from .._forms import ShapeFormHandler
+from .._interfaces import OptimizationProblem
 from .._loggers import debug, warning
 from .._pde_problems import AdjointProblem, ShapeGradientProblem, StateProblem
 from .._shape_optimization import ReducedShapeCostFunctional
 from ..geometry import _MeshHandler
-from ..optimization_problem import OptimizationProblem
-from ..utils import _optimization_algorithm_configuration
 
 
 class ShapeOptimizationProblem(OptimizationProblem):
@@ -55,73 +61,157 @@ class ShapeOptimizationProblem(OptimizationProblem):
     and so on.
     """
 
+    def __new__(
+        cls,
+        state_forms: Union[ufl.Form, List[ufl.Form]],
+        bcs_list: Union[
+            fenics.DirichletBC,
+            List[fenics.DirichletBC],
+            List[List[fenics.DirichletBC]],
+            None,
+        ],
+        cost_functional_form: Union[ufl.Form, List[ufl.Form]],
+        states: Union[fenics.Function, List[fenics.Function]],
+        adjoints: Union[fenics.Function, List[fenics.Function]],
+        boundaries: fenics.MeshFunction,
+        config: Optional[configparser.ConfigParser] = None,
+        shape_scalar_product: Optional[ufl.Form] = None,
+        initial_guess: Optional[List[fenics.Function]] = None,
+        ksp_options: Optional[List[List[List[str]]]] = None,
+        adjoint_ksp_options: Optional[List[List[List[str]]]] = None,
+        scalar_tracking_forms: Optional[Dict] = None,
+        min_max_terms: Optional[Dict] = None,
+        desired_weights: Optional[List[float]] = None,
+    ) -> ShapeOptimizationProblem:
+
+        if desired_weights is not None:
+            _use_scaling = True
+        else:
+            _use_scaling = False
+
+        if _use_scaling:
+            unscaled_problem = super().__new__(cls)
+            unscaled_problem.__init__(
+                state_forms,
+                bcs_list,
+                cost_functional_form,
+                states,
+                adjoints,
+                boundaries,
+                config=config,
+                shape_scalar_product=shape_scalar_product,
+                initial_guess=initial_guess,
+                ksp_options=ksp_options,
+                adjoint_ksp_options=adjoint_ksp_options,
+                scalar_tracking_forms=scalar_tracking_forms,
+                min_max_terms=min_max_terms,
+                desired_weights=desired_weights,
+            )
+            unscaled_problem._scale_cost_functional()  # overwrites the cost functional list
+
+            problem = super().__new__(cls)
+            if not unscaled_problem.has_cashocs_remesh_flag:
+                problem.initial_function_values = (
+                    unscaled_problem.initial_function_values
+                )
+                if unscaled_problem.use_scalar_tracking:
+                    problem.initial_scalar_tracking_values = (
+                        unscaled_problem.initial_scalar_tracking_values
+                    )
+
+            if (
+                not unscaled_problem.has_cashocs_remesh_flag
+                and unscaled_problem.do_remesh
+            ):
+                subprocess.run(["rm", "-r", unscaled_problem.temp_dir], check=True)
+                subprocess.run(
+                    ["rm", "-r", unscaled_problem.mesh_handler.remesh_directory],
+                    check=True,
+                )
+
+            return problem
+
+        else:
+            return super().__new__(cls)
+
     def __init__(
         self,
-        state_forms,
-        bcs_list,
-        cost_functional_form,
-        states,
-        adjoints,
-        boundaries,
-        config=None,
-        shape_scalar_product=None,
-        initial_guess=None,
-        ksp_options=None,
-        adjoint_ksp_options=None,
-        desired_weights=None,
-        scalar_tracking_forms=None,
-    ):
-        """This is used to generate all classes and functionalities. First ensures
-        consistent input, afterwards, the solution algorithm is initialized.
-
+        state_forms: Union[ufl.Form, List[ufl.Form]],
+        bcs_list: Union[
+            fenics.DirichletBC,
+            List[fenics.DirichletBC],
+            List[List[fenics.DirichletBC]],
+            None,
+        ],
+        cost_functional_form: Union[ufl.Form, List[ufl.Form]],
+        states: Union[fenics.Function, List[fenics.Function]],
+        adjoints: Union[fenics.Function, List[fenics.Function]],
+        boundaries: fenics.MeshFunction,
+        config: Optional[configparser.ConfigParser] = None,
+        shape_scalar_product: Optional[ufl.Form] = None,
+        initial_guess: Optional[List[fenics.Function]] = None,
+        ksp_options: Optional[List[List[List[str]]]] = None,
+        adjoint_ksp_options: Optional[List[List[List[str]]]] = None,
+        scalar_tracking_forms: Optional[Dict] = None,
+        min_max_terms: Optional[Dict] = None,
+        desired_weights: Optional[List[float]] = None,
+    ) -> None:
+        """
         Parameters
         ----------
-        state_forms : ufl.form.Form or list[ufl.form.Form]
-                The weak form of the state equation (user implemented). Can be either
-                a single UFL form, or a (ordered) list of UFL forms.
-        bcs_list : list[dolfin.fem.dirichletbc.DirichletBC] or list[list[dolfin.fem.dirichletbc.DirichletBC]] or dolfin.fem.dirichletbc.DirichletBC or None
-                The list of DirichletBC objects describing Dirichlet (essential) boundary conditions.
-                If this is ``None``, then no Dirichlet boundary conditions are imposed.
-        cost_functional_form : ufl.form.Form or list[ufl.form.Form]
-                UFL form of the cost functional.
-        states : dolfin.function.function.Function or list[dolfin.function.function.Function]
-                The state variable(s), can either be a :py:class:`fenics.Function`, or a list of these.
-        adjoints : dolfin.function.function.Function or list[dolfin.function.function.Function]
-                The adjoint variable(s), can either be a :py:class:`fenics.Function`, or a (ordered) list of these.
-        boundaries : dolfin.cpp.mesh.MeshFunctionSizet.MeshFunctionSizet
-                :py:class:`fenics.MeshFunction` that indicates the boundary markers.
-        config : configparser.ConfigParser or None
-                The config file for the problem, generated via :py:func:`cashocs.create_config`.
-                Alternatively, this can also be ``None``, in which case the default configurations
-                are used, except for the optimization algorithm. This has then to be specified
-                in the :py:meth:`solve <cashocs.OptimalControlProblem.solve>` method. The
-                default is ``None``.
-        shape_scalar_product : ufl.form.Form
-                The bilinear form for computing the shape gradient (or gradient deformation).
-                This has to use :py:class:`fenics.TrialFunction` and :py:class:`fenics.TestFunction`
-                objects to define the weak form, which have to be in a :py:class:`fenics.VectorFunctionSpace`
-                of continuous, linear Lagrange finite elements. Moreover, this form is required to be
-                symmetric.
-        initial_guess : list[dolfin.function.function.Function], optional
-                List of functions that act as initial guess for the state variables, should be valid input for :py:func:`fenics.assign`.
-                Defaults to ``None``, which means a zero initial guess.
+        state_forms : ufl.Form or list[ufl.Form]
+            The weak form of the state equation (user implemented). Can be either
+            a single UFL form, or a (ordered) list of UFL forms.
+        bcs_list : list[fenics.DirichletBC] or list[list[fenics.DirichletBC]] or fenics.DirichletBC or None
+            The list of DirichletBC objects describing Dirichlet (essential) boundary conditions.
+            If this is ``None``, then no Dirichlet boundary conditions are imposed.
+        cost_functional_form : ufl.Form or list[ufl.Form]
+            UFL form of the cost functional.
+        states : fenics.Function or list[fenics.Function]
+            The state variable(s), can either be a :py:class:`fenics.Function`, or a list of these.
+        adjoints : fenics.Function or list[fenics.Function]
+            The adjoint variable(s), can either be a :py:class:`fenics.Function`, or a (ordered) list of these.
+        boundaries : fenics.MeshFunction
+            :py:class:`fenics.MeshFunction` that indicates the boundary markers.
+        config : configparser.ConfigParser or None, optional
+            The config file for the problem, generated via :py:func:`cashocs.create_config`.
+            Alternatively, this can also be ``None``, in which case the default configurations
+            are used, except for the optimization algorithm. This has then to be specified
+            in the :py:meth:`solve <cashocs.OptimalControlProblem.solve>` method. The
+            default is ``None``.
+        shape_scalar_product : ufl.form.Form or None, optional
+            The bilinear form for computing the shape gradient (or gradient deformation).
+            This has to use :py:class:`fenics.TrialFunction` and :py:class:`fenics.TestFunction`
+            objects to define the weak form, which have to be in a :py:class:`fenics.VectorFunctionSpace`
+            of continuous, linear Lagrange finite elements. Moreover, this form is required to be
+            symmetric.
+        initial_guess : list[fenics.Function] or None, optional
+            List of functions that act as initial guess for the state variables, should be valid input for :py:func:`fenics.assign`.
+            Defaults to ``None``, which means a zero initial guess.
         ksp_options : list[list[str]] or list[list[list[str]]] or None, optional
-                A list of strings corresponding to command line options for PETSc,
-                used to solve the state systems. If this is ``None``, then the direct solver
-                mumps is used (default is ``None``).
+            A list of strings corresponding to command line options for PETSc,
+            used to solve the state systems. If this is ``None``, then the direct solver
+            mumps is used (default is ``None``).
         adjoint_ksp_options : list[list[str]] or list[list[list[str]]] or None
-                A list of strings corresponding to command line options for PETSc,
-                used to solve the adjoint systems. If this is ``None``, then the same options
-                as for the state systems are used (default is ``None``).
-        scalar_tracking_forms : dict or list[dict]
-                A list of dictionaries that define scalar tracking type cost functionals,
-                where an integral value should be brought to a desired value. Each dict needs
-                to have the keys ``'integrand'`` and ``'tracking_goal'``. Default is ``None``,
-                i.e., no scalar tracking terms are considered.
+            A list of strings corresponding to command line options for PETSc,
+            used to solve the adjoint systems. If this is ``None``, then the same options
+            as for the state systems are used (default is ``None``).
+        scalar_tracking_forms : dict or list[dict] or None, optional
+            A list of dictionaries that define scalar tracking type cost functionals,
+            where an integral value should be brought to a desired value. Each dict needs
+            to have the keys ``'integrand'`` and ``'tracking_goal'``. Default is ``None``,
+            i.e., no scalar tracking terms are considered.
+        min_max_terms : dict or list[dict] or None, optional
+            Additional terms for the cost functional, not to be used directly.
+        desired_weights : list[float] or None, optional
+            A list of values for scaling the cost functional terms. If this is supplied,
+            the cost functional has to be given as list of summands. The individual terms
+            are then scaled, so that term `i` has the magnitude of `desired_weights[i]`
+            for the initial iteration. In case that `desired_weights` is `None`, no scaling
+            is performed. Default is `None`.
         """
 
-        OptimizationProblem.__init__(
-            self,
+        super().__init__(
             state_forms,
             bcs_list,
             cost_functional_form,
@@ -131,8 +221,9 @@ class ShapeOptimizationProblem(OptimizationProblem):
             initial_guess,
             ksp_options,
             adjoint_ksp_options,
-            desired_weights,
             scalar_tracking_forms,
+            min_max_terms,
+            desired_weights,
         )
 
         ### Initialize the remeshing behavior, and a temp file
@@ -154,10 +245,7 @@ class ShapeOptimizationProblem(OptimizationProblem):
                 pass
 
             try:
-                if (
-                    not self.states[0].function_space().mesh()._cashocs_generator
-                    == "config"
-                ):
+                if not self.states[0].function_space().mesh()._config_flag:
                     raise InputError(
                         "cashocs.import_mesh",
                         "arg",
@@ -187,14 +275,17 @@ class ShapeOptimizationProblem(OptimizationProblem):
                     "output_dict": {},
                 }
 
-                if self.bool_scaling:
-                    self.temp_dict[
-                        "initial_function_values"
-                    ] = self.initial_function_values
-                    if self.use_scalar_tracking:
+                try:
+                    if self.use_scaling:
                         self.temp_dict[
-                            "intitial_scalar_tracking_values"
-                        ] = self.initial_scalar_tracking_values
+                            "initial_function_values"
+                        ] = self.initial_function_values
+                        if self.use_scalar_tracking:
+                            self.temp_dict[
+                                "initial_scalar_tracking_values"
+                            ] = self.initial_scalar_tracking_values
+                except AttributeError:
+                    pass
 
             else:
                 self.__change_except_hook()
@@ -212,66 +303,33 @@ class ShapeOptimizationProblem(OptimizationProblem):
             )
 
         # shape_scalar_product
+        self.shape_scalar_product = shape_scalar_product
         if shape_scalar_product is None:
-            self.shape_scalar_product = shape_scalar_product
             self.deformation_space = None
-
         else:
-            if isinstance(shape_scalar_product, ufl.form.Form):
-                if len(shape_scalar_product.arguments()) == 2:
-                    family = (
-                        shape_scalar_product.arguments()[0]
-                        .ufl_function_space()
-                        .ufl_element()
-                        .family()
-                    )
-                    degree = (
-                        shape_scalar_product.arguments()[0]
-                        .ufl_function_space()
-                        .ufl_element()
-                        .degree()
-                    )
-                    space_0 = shape_scalar_product.arguments()[0].ufl_function_space()
-                    space_1 = shape_scalar_product.arguments()[1].ufl_function_space()
+            self.deformation_space = self.shape_scalar_product.arguments()[
+                0
+            ].ufl_function_space()
 
-                    if family == "Lagrange" and degree == 1 and space_0 == space_1:
-                        self.shape_scalar_product = shape_scalar_product
-                        self.deformation_space = space_0
-                else:
-                    raise InputError(
-                        "cashocs._shape_optimization.shape_optimization_problem.ShapeOptimizationProblem",
-                        "shape_scalar_product",
-                        "The supplied form must be a bilinear one.",
-                    )
-            else:
-                raise InputError(
-                    "cashocs._shape_optimization.shape_optimization_problem.ShapeOptimizationProblem",
-                    "shape_scalar_product",
-                    "Not a valid type for shape_scalar_product.",
-                )
+        if self.shape_scalar_product is not None:
+            self.uses_custom_scalar_product = True
 
-        self.form_handler = ShapeFormHandler(
-            self.lagrangian,
-            self.bcs_list,
-            self.states,
-            self.adjoints,
-            self.boundaries,
-            self.config,
-            self.ksp_options,
-            self.adjoint_ksp_options,
-            self.shape_scalar_product,
-            self.deformation_space,
-            use_scalar_tracking=self.use_scalar_tracking,
-            has_cashocs_remesh_flag=self.has_cashocs_remesh_flag,
-            temp_dir=self.temp_dir,
-        )
+        if self.uses_custom_scalar_product and self.config.getboolean(
+            "ShapeGradient", "use_p_laplacian", fallback=False
+        ):
+            warning(
+                "You have supplied a custom scalar product and set the parameter ``use_p_laplacian`` in the config file."
+                "cashocs will use the supplied scalar product and not the p-Laplacian to compute the shape gradient."
+            )
+
+        self.form_handler = ShapeFormHandler(self)
 
         if self.do_remesh and not self.has_cashocs_remesh_flag:
             self.temp_dict["Regularization"] = {
-                "mu_volume": self.form_handler.regularization.mu_volume,
-                "mu_surface": self.form_handler.regularization.mu_surface,
-                "mu_curvature": self.form_handler.regularization.mu_curvature,
-                "mu_barycenter": self.form_handler.regularization.mu_barycenter,
+                "mu_volume": self.form_handler.shape_regularization.mu_volume,
+                "mu_surface": self.form_handler.shape_regularization.mu_surface,
+                "mu_curvature": self.form_handler.shape_regularization.mu_curvature,
+                "mu_barycenter": self.form_handler.shape_regularization.mu_barycenter,
             }
 
         self.mesh_handler = _MeshHandler(self)
@@ -285,7 +343,7 @@ class ShapeOptimizationProblem(OptimizationProblem):
         self.adjoint_problem = AdjointProblem(
             self.form_handler, self.state_problem, self.temp_dict
         )
-        self.shape_gradient_problem = ShapeGradientProblem(
+        self.gradient_problem = ShapeGradientProblem(
             self.form_handler, self.state_problem, self.adjoint_problem
         )
 
@@ -293,10 +351,10 @@ class ShapeOptimizationProblem(OptimizationProblem):
             self.form_handler, self.state_problem
         )
 
-        self.gradient = self.shape_gradient_problem.gradient
+        self.gradient = self.gradient_problem.gradient
         self.objective_value = 1.0
 
-    def _erase_pde_memory(self):
+    def _erase_pde_memory(self) -> None:
         """Resets the memory of the PDE problems so that new solutions are computed.
 
         This sets the value of has_solution to False for all relevant PDE problems,
@@ -307,40 +365,55 @@ class ShapeOptimizationProblem(OptimizationProblem):
         None
         """
 
+        super()._erase_pde_memory()
         self.mesh_handler.bbtree.build(self.mesh_handler.mesh)
         self.form_handler.update_scalar_product()
-        self.state_problem.has_solution = False
-        self.adjoint_problem.has_solution = False
-        self.shape_gradient_problem.has_solution = False
+        self.gradient_problem.has_solution = False
 
-    def solve(self, algorithm=None, rtol=None, atol=None, max_iter=None):
+    def solve(
+        self,
+        algorithm: Optional[
+            Literal[
+                "gradient_descent",
+                "gd",
+                "conjugate_gradient",
+                "nonlinear_cg",
+                "ncg",
+                "lbfgs",
+                "bfgs",
+            ]
+        ] = None,
+        rtol: Optional[float] = None,
+        atol: Optional[float] = None,
+        max_iter: Optional[int] = None,
+    ) -> None:
         r"""Solves the optimization problem by the method specified in the config file.
 
         Parameters
         ----------
         algorithm : str or None, optional
-                Selects the optimization algorithm. Valid choices are
-                ``'gradient_descent'`` or ``'gd'`` for a gradient descent method,
-                ``'conjugate_gradient'``, ``'nonlinear_cg'``, ``'ncg'`` or ``'cg'``
-                for nonlinear conjugate gradient methods, and ``'lbfgs'`` or ``'bfgs'`` for
-                limited memory BFGS methods. This overwrites the value specified
-                in the config file. If this is ``None``, then the value in the
-                config file is used. Default is ``None``.
+            Selects the optimization algorithm. Valid choices are
+            ``'gradient_descent'`` or ``'gd'`` for a gradient descent method,
+            ``'conjugate_gradient'``, ``'nonlinear_cg'``, ``'ncg'`` or ``'cg'``
+            for nonlinear conjugate gradient methods, and ``'lbfgs'`` or ``'bfgs'`` for
+            limited memory BFGS methods. This overwrites the value specified
+            in the config file. If this is ``None``, then the value in the
+            config file is used. Default is ``None``.
         rtol : float or None, optional
-                The relative tolerance used for the termination criterion.
-                Overwrites the value specified in the config file. If this
-                is ``None``, the value from the config file is taken. Default
-                is ``None``.
+            The relative tolerance used for the termination criterion.
+            Overwrites the value specified in the config file. If this
+            is ``None``, the value from the config file is taken. Default
+            is ``None``.
         atol : float or None, optional
-                The absolute tolerance used for the termination criterion.
-                Overwrites the value specified in the config file. If this
-                is ``None``, the value from the config file is taken. Default
-                is ``None``.
+            The absolute tolerance used for the termination criterion.
+            Overwrites the value specified in the config file. If this
+            is ``None``, the value from the config file is taken. Default
+            is ``None``.
         max_iter : int or None, optional
-                The maximum number of iterations the optimization algorithm
-                can carry out before it is terminated. Overwrites the value
-                specified in the config file. If this is ``None``, the value from
-                the config file is taken. Default is ``None``.
+            The maximum number of iterations the optimization algorithm
+            can carry out before it is terminated. Overwrites the value
+            specified in the config file. If this is ``None``, the value from
+            the config file is taken. Default is ``None``.
 
         Returns
         -------
@@ -364,22 +437,7 @@ class ShapeOptimizationProblem(OptimizationProblem):
           .. math:: || \nabla J(u_k) || \leq \texttt{atol} + \texttt{rtol} || \nabla J(u_0) ||
         """
 
-        self.algorithm = _optimization_algorithm_configuration(self.config, algorithm)
-
-        if (rtol is not None) and (atol is None):
-            self.config.set("OptimizationRoutine", "rtol", str(rtol))
-            self.config.set("OptimizationRoutine", "atol", str(0.0))
-        elif (atol is not None) and (rtol is None):
-            self.config.set("OptimizationRoutine", "rtol", str(0.0))
-            self.config.set("OptimizationRoutine", "atol", str(atol))
-        elif (atol is not None) and (rtol is not None):
-            self.config.set("OptimizationRoutine", "rtol", str(rtol))
-            self.config.set("OptimizationRoutine", "atol", str(atol))
-
-        if max_iter is not None:
-            self.config.set("OptimizationRoutine", "maximum_iterations", str(max_iter))
-
-        self._check_for_custom_forms()
+        super().solve(algorithm=algorithm, rtol=rtol, atol=atol, max_iter=max_iter)
 
         if self.algorithm == "gradient_descent":
             self.solver = GradientDescent(self)
@@ -395,18 +453,11 @@ class ShapeOptimizationProblem(OptimizationProblem):
                 "method. Needs to be one of 'gradient_descent' ('gd'), 'lbfgs' ('bfgs'), "
                 "or 'conjugate_gradient' ('cg').",
             )
-        else:
-            raise ConfigError(
-                "OptimizationRoutine",
-                "algorithm",
-                "Not a valid input. Needs to be one "
-                "of 'gradient_descent' ('gd'), 'lbfgs' ('bfgs'), or 'conjugate_gradient' ('cg').",
-            )
 
         self.solver.run()
         self.solver.post_processing()
 
-    def __change_except_hook(self):
+    def __change_except_hook(self) -> None:
         """Ensures that temp files are deleted when an exception occurs.
 
         This modifies the sys.excepthook command so that it also deletes temp files
@@ -418,6 +469,18 @@ class ShapeOptimizationProblem(OptimizationProblem):
         """
 
         def custom_except_hook(exctype, value, traceback):
+            """A customized hook which is injected when an exception occurs.
+
+            Parameters
+            ----------
+            exctype
+            value
+            traceback
+
+            Returns
+            -------
+
+            """
             debug(
                 "An exception was raised by cashocs, deleting the created temporary files."
             )
@@ -430,7 +493,7 @@ class ShapeOptimizationProblem(OptimizationProblem):
 
         sys.excepthook = custom_except_hook
 
-    def compute_shape_gradient(self):
+    def compute_shape_gradient(self) -> fenics.Function:
         """Solves the Riesz problem to determine the shape gradient.
 
         This can be used for debugging, or code validation.
@@ -439,15 +502,15 @@ class ShapeOptimizationProblem(OptimizationProblem):
 
         Returns
         -------
-        dolfin.function.function.Function
-                The shape gradient.
+        fenics.Function
+            The shape gradient.
         """
 
-        self.shape_gradient_problem.solve()
+        self.gradient_problem.solve()
 
         return self.gradient
 
-    def supply_shape_derivative(self, shape_derivative):
+    def supply_shape_derivative(self, shape_derivative: ufl.Form) -> None:
         """Overrides the shape derivative of the reduced cost functional.
 
         This allows users to implement their own shape derivative and use cashocs as a
@@ -455,8 +518,8 @@ class ShapeOptimizationProblem(OptimizationProblem):
 
         Parameters
         ----------
-        shape_derivative : ufl.form.Form
-                The shape_derivative of the reduced (!) cost functional w.r.t. controls.
+        shape_derivative : ufl.Form
+            The shape_derivative of the reduced (!) cost functional w.r.t. controls.
 
         Returns
         -------
@@ -545,7 +608,14 @@ class ShapeOptimizationProblem(OptimizationProblem):
 
         self.has_custom_derivative = True
 
-    def supply_custom_forms(self, shape_derivative, adjoint_forms, adjoint_bcs_list):
+    def supply_custom_forms(
+        self,
+        shape_derivative: ufl.Form,
+        adjoint_forms: Union[ufl.Form, List[ufl.Form]],
+        adjoint_bcs_list: Union[
+            fenics.DirichletBC, List[fenics.DirichletBC], List[List[fenics.DirichletBC]]
+        ],
+    ) -> None:
         """Overrides both adjoint system and shape derivative with user input.
 
         This allows the user to specify both the shape_derivative of the reduced cost functional
@@ -558,12 +628,12 @@ class ShapeOptimizationProblem(OptimizationProblem):
 
         Parameters
         ----------
-        shape_derivative : ufl.form.Form
-                The shape derivative of the reduced (!) cost functional.
-        adjoint_forms : ufl.form.Form or list[ufl.form.Form]
-                The UFL forms of the adjoint system(s).
-        adjoint_bcs_list : list[dolfin.fem.dirichletbc.DirichletBC] or list[list[dolfin.fem.dirichletbc.DirichletBC]] or dolfin.fem.dirichletbc.DirichletBC or None
-                The list of Dirichlet boundary conditions for the adjoint system(s).
+        shape_derivative : ufl.Form
+            The shape derivative of the reduced (!) cost functional.
+        adjoint_forms : ufl.Form or list[ufl.Form]
+            The UFL forms of the adjoint system(s).
+        adjoint_bcs_list : list[fenics.DirichletBC] or list[list[fenics.DirichletBC]] or fenics.DirichletBC or None
+            The list of Dirichlet boundary conditions for the adjoint system(s).
 
         Returns
         -------
@@ -573,7 +643,7 @@ class ShapeOptimizationProblem(OptimizationProblem):
         self.supply_shape_derivative(shape_derivative)
         self.supply_adjoint_forms(adjoint_forms, adjoint_bcs_list)
 
-    def get_vector_field(self):
+    def get_vector_field(self) -> dolfin.function.argument.Argument:
         """Returns the TestFunction for defining shape derivatives.
 
         See Also
@@ -582,8 +652,32 @@ class ShapeOptimizationProblem(OptimizationProblem):
 
         Returns
         -------
-         : dolfin.function.argument.Argument
-                The TestFunction object.
+        dolfin.function.argument.Argument
+            The TestFunction object.
         """
 
         return self.form_handler.test_vector_field
+
+    def gradient_test(
+        self,
+        h: Optional[fenics.Function] = None,
+        rng: Optional[np.random.RandomState] = None,
+    ) -> float:
+        """Taylor test to verify that the computed shape gradient is correct.
+
+        Parameters
+        ----------
+        h : fenics.Function or None, optional
+            The direction used to compute the directional derivative. If this is
+            ``None``, then a random direction is used (default is ``None``).
+        rng : numpy.random.RandomState or None, optional
+            A numpy random state for calculating a random direction
+
+        Returns
+        -------
+         float
+            The convergence order from the Taylor test. If this is (approximately) 2 or larger,
+            everything works as expected.
+        """
+
+        return verification.shape_gradient_test(self, h, rng)
